/* render.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include "matrix.h"
#include "globals.h"
#include "graphics.h"
#include "labels.h"
#include "map.h"
#include "memory.h"
#include "queue.h"
#include "sync.h"
#include "topo.h"
#include "x.h"


#define MAX(A,B)  ( (A) > (B) ? (A) : (B) )
#define MIN(A,B)  ( (A) < (B) ? (A) : (B) )


static float viewbox[24][3], arrow[6][3], N[4][3], ticks[12][3];



void draw_box(), draw_surfaces(), draw_hslices(), draw_vslices(),
  draw_clock(), draw_chslices(), draw_cvslices(), plot_string(),
  draw_slice_tick(), draw_hwind(), draw_vwind(), draw_cursor(),
  draw_traj(), draw_info();




/*** update_display ***************************************************
   Redraw the contents of the 3-D window.
   Input:  none.
   Output:  none.
**********************************************************************/
void update_display()
{
   int labels, i;

   if (Debug)
     printf("================== update display =================\n");
   if (GoTime)
      labels = !ContnumFlag;
   else
      labels = ContnumFlag;

   set_3d( CTM );

   /* Loop over antialiasing passes */
   for (i=0; i < (PrettyFlag ? AA_PASSES : 1); i++) {
      /*printf("pass: %d\n", i );*/

      start_aa_pass(i);
      clear_window( BgColor );

      /*** Draw 3-D lines ***/

      draw_traj( CurTime );

      if (DisplayBox)
	draw_box( CurTime );

      if (DisplayCursor)
	draw_cursor( CurTime, DisplayBox );

      get_user_input();

      draw_hslices( CurTime, labels );
      draw_vslices( CurTime, labels );

      get_user_input();

      draw_hwind( CurTime );
      draw_vwind( CurTime );

      /*** Draw opaque 3-D surfaces ***/

      if (TopoFlag && DisplayTopo) {
	 draw_topo( CurTime );
	 if (MapFlag && DisplayMap) {
	    set_color( BoxColor );
	    draw_map( CurTime, 0xff000000, 0 );
	 }
      }
      else if (MapFlag && DisplayMap) {
	 draw_map( CurTime, BoxColor, 1 );
      }
      get_user_input();

      draw_surfaces( CurTime, 1 );
      get_user_input();

      draw_chslices( CurTime, 1 );
      get_user_input();

      draw_cvslices( CurTime, 1 );
      get_user_input();

      /* Draw transparent 3-D objects */
      if (PrettyFlag) {
	 /* multi-pass transparency */
	 do {
	    start_hqtrans_pass();
	    draw_surfaces( CurTime, 0 );
	    get_user_input();
	    draw_chslices( CurTime, 0 );
	    get_user_input();
	    draw_cvslices( CurTime, 0 );
	    get_user_input();
	 } while (end_hqtrans_pass() > 200);
      }
      else {
	 /* single-pass transparency */
	 draw_surfaces( CurTime, 0 );
	 get_user_input();
	 draw_chslices( CurTime, 0 );
	 get_user_input();
	 draw_cvslices( CurTime, 0 );
	 get_user_input();
      }

      /*** Draw 2-D objects ***/
      set_2d();
      if (DisplayClock) {
	 draw_clock( BoxColor );
	 draw_logo( BoxColor );
      }
      if (DisplayInfo) {
	 draw_info();
      }

      draw_labels();
      get_user_input();

      end_aa_pass(i);

   } /* aa passes */

   update_window();
}





/*** init_box *********************************************************
   Initialize the graphics arrays for the 3-D viewbox.
   This function must be called AFTER the Nr,Nc,Nl variables have
   been initialized.
   Input:  prop - code for box proportions:
                  1 = scale box according to lat, lons.
                  2 = scale box according to grid size.
                  3 = scale according to ax, ay, az.
	   ax, ay, az - aspect ratio numbers used if prop==3.
   Output:  The global vars Gxscale, Gyscale, Gzscale, and the
            viewbox are initialized.
**********************************************************************/
void init_box( prop, ax, ay, az )
int prop;
float ax, ay, az;
{
   float lats, lons, hgt, midlat;

   if (prop==1) {
      /* scale box according to lat and lon */

      midlat = (NorthLat[0]+SouthLat[0]) / 2.0;
      lats = NorthLat[0]-SouthLat[0];
      lons = (WestLon[0]-EastLon[0]) * cos( midlat * 3.14159/180.0 );
      hgt = TopHgt[0]-BottomHgt[0];

      if (lons>lats) {
	 Gxscale = 1.0;
	 Gyscale = lats / lons;
	 Gzscale = Gyscale / 2.0;
      }
      else {
	 Gxscale = lons/ lats;
	 Gyscale = 1.0;
	 Gzscale = Gxscale / 2.0;
      }

      /* make sure proportions are reasonable */
      if (Gxscale<0.3)
	 Gxscale = 0.3;
      if (Gyscale<0.3)
	 Gyscale = 0.3;
      if (Gzscale<0.3)
	 Gzscale = 0.3;
   }
   else if (prop==2) {
      /* scale box according to grid points */
      if (Nc>=Nr && Nc>=Nl) {
	 Gxscale = 1.0;
	 Gyscale = (float) Nr / (float) Nc;
	 Gzscale = (float) Nl / (float) Nc;
      }
      else if (Nr>=Nc && Nr>=Nl) {
	 Gxscale = (float) Nc / (float) Nr;
	 Gyscale = 1.0;
	 Gzscale = (float) Nl / (float) Nr;
      }
      else {
	 Gxscale = (float) Nc / (float) Nl;
	 Gyscale = (float) Nr / (float) Nl;
	 Gzscale = 1.0;
      }
   }
   else if (prop==3) {
      /* scale box according to aspect ratios */
      if (ax>=ay && ax>=az) {
	 ay = ay / ax;
	 az = az / ax;
	 ax = 1.0;
      }
      else if (ay>=ax && ay>=az) {
	 ax = ax / ay;
	 az = az / ay;
	 ay = 1.0;
      }
      else {
	 ax = ax / az;
	 ay = ay / az;
	 az = 1.0;
      }
      Gxscale = ax;
      Gyscale = ay;
      Gzscale = az;
   }

   Xmin = -Gxscale;
   Xmax = Gxscale;
   Ymin = -Gyscale;
   Ymax = Gyscale;
   Zmin = -Gzscale;
   Zmax = Gzscale;

   viewbox[0][0] = Xmax;   viewbox[0][1] = Ymin;   viewbox[0][2] = Zmax;
   viewbox[1][0] = Xmax;   viewbox[1][1] = Ymax;   viewbox[1][2] = Zmax;
   viewbox[2][0] = Xmax;   viewbox[2][1] = Ymax;   viewbox[2][2] = Zmax;
   viewbox[3][0] = Xmin;   viewbox[3][1] = Ymax;   viewbox[3][2] = Zmax;
   viewbox[4][0] = Xmin;   viewbox[4][1] = Ymin;   viewbox[4][2] = Zmin;
   viewbox[5][0] = Xmax;   viewbox[5][1] = Ymin;   viewbox[5][2] = Zmin;
   viewbox[6][0] = Xmax;   viewbox[6][1] = Ymin;   viewbox[6][2] = Zmin;
   viewbox[7][0] = Xmax;   viewbox[7][1] = Ymax;   viewbox[7][2] = Zmin;
   viewbox[8][0] = Xmax;   viewbox[8][1] = Ymax;   viewbox[8][2] = Zmin;
   viewbox[9][0] = Xmin;   viewbox[9][1] = Ymax;   viewbox[9][2] = Zmin;
   viewbox[10][0] = Xmin;  viewbox[10][1] = Ymax;  viewbox[10][2] = Zmin;
   viewbox[11][0] = Xmin;  viewbox[11][1] = Ymin;  viewbox[11][2] = Zmin;
   viewbox[12][0] = Xmax;  viewbox[12][1] = Ymin;  viewbox[12][2] = Zmax;
   viewbox[13][0] = Xmax;  viewbox[13][1] = Ymin;  viewbox[13][2] = Zmin;
   viewbox[14][0] = Xmax;  viewbox[14][1] = Ymax;  viewbox[14][2] = Zmax;
   viewbox[15][0] = Xmax;  viewbox[15][1] = Ymax;  viewbox[15][2] = Zmin;
   viewbox[16][0] = Xmin;  viewbox[16][1] = Ymax;  viewbox[16][2] = Zmax;
   viewbox[17][0] = Xmin;  viewbox[17][1] = Ymax;  viewbox[17][2] = Zmin;
   viewbox[18][0] = Xmin;  viewbox[18][1] = Ymin;  viewbox[18][2] = Zmax;
   viewbox[19][0] = Xmax;  viewbox[19][1] = Ymin;  viewbox[19][2] = Zmax;
   viewbox[20][0] = Xmin;  viewbox[20][1] = Ymin;  viewbox[20][2] = Zmax;
   viewbox[21][0] = Xmin;  viewbox[21][1] = Ymax;  viewbox[21][2] = Zmax;
   viewbox[22][0] = Xmin;  viewbox[22][1] = Ymin;  viewbox[22][2] = Zmax;
   viewbox[23][0] = Xmin;  viewbox[23][1] = Ymin;  viewbox[23][2] = Zmin;

   arrow[0][0] =  0.0;    arrow[0][1] = Ymax+0.1;   arrow[0][2] = Zmax;
   arrow[1][0] = -0.025;  arrow[1][1] = Ymax+0.05;  arrow[1][2] = Zmax;
   arrow[2][0] =  0.0;    arrow[2][1] = Ymax+0.1;   arrow[2][2] = Zmax;
   arrow[3][0] =  0.0;    arrow[3][1] = Ymax;       arrow[3][2] = Zmax;
   arrow[4][0] =  0.0;    arrow[4][1] = Ymax+0.1;   arrow[4][2] = Zmax;
   arrow[5][0] =  0.025;  arrow[5][1] = Ymax+0.05;  arrow[5][2] = Zmax;

   N[0][0] = -0.025;  N[0][1] = Ymax+0.15;  N[0][2] = Zmax;
   N[1][0] = -0.025;  N[1][1] = Ymax+0.25;  N[1][2] = Zmax;
   N[2][0] =  0.025;  N[2][1] = Ymax+0.15;  N[2][2] = Zmax;
   N[3][0] =  0.025;  N[3][1] = Ymax+0.25;  N[3][2] = Zmax;

   /* East */
   ticks[0][0] = Xmax;  ticks[0][1] = Ymin;       ticks[0][2] = Zmax;
   ticks[1][0] = Xmax;  ticks[1][1] = Ymin-0.05;  ticks[1][2] = Zmax+0.05;
   /* West */
   ticks[2][0] = Xmin;  ticks[2][1] = Ymin;       ticks[2][2] = Zmax;
   ticks[3][0] = Xmin;  ticks[3][1] = Ymin-0.05;  ticks[3][2] = Zmax+0.05;
   /* North */
   ticks[4][0] = Xmin;       ticks[4][1] = Ymax;  ticks[4][2] = Zmax;
   ticks[5][0] = Xmin-0.05;  ticks[5][1] = Ymax;  ticks[5][2] = Zmax+0.05;
   /* South */
   ticks[6][0] = Xmin;       ticks[6][1] = Ymin;  ticks[6][2] = Zmax;
   ticks[7][0] = Xmin-0.05;  ticks[7][1] = Ymin;  ticks[7][2] = Zmax+0.05;
   /* Top */
   ticks[8][0] = Xmin;       ticks[8][1] = Ymin;       ticks[8][2] = Zmin;
   ticks[9][0] = Xmin-0.05;  ticks[9][1] = Ymin-0.05;  ticks[9][2] = Zmin;
   /* Bottom */
   ticks[10][0] = Xmin;       ticks[10][1] = Ymin;       ticks[10][2] = Zmax;
   ticks[11][0] = Xmin-0.05;  ticks[11][1] = Ymin-0.05;  ticks[11][2] = Zmax;
}



/*** draw_box *********************************************************
   Draw the 3-D view box.
   Input:  it - time step.
   Output:  none.
**********************************************************************/
void draw_box( it )
int it;
{
   /* base and up vectors for text drawn along x,y,z axes. */
   static float bx[3] = { 0.05, 0.0, 0.0 },      ux[3] = { 0.0, 0.05, -0.05 };
   static float by[3] = { -0.035, 0.0, 0.035 },  uy[3] = { 0.0, 0.07, 0.0 };
   static float bz[3] = { -0.035, -0.035, 0.0 }, uz[3] = { 0.0, 0.0, -0.07 };
   float x1, y1, z1, x2, y2, z2;
   char str[100];

   /* set depth cueing & line color */
   set_color( BoxColor );
   set_depthcue( 1 );

   /* Draw box. */
   disjointpolyline( viewbox, 24 );

   /* Draw north-pointing arrow and 'N'. */
   disjointpolyline( arrow, 6 );
   polyline( N, 4 );

   /* tick marks */
   disjointpolyline( ticks, 12 );

   /* Draw axis labels. */
   if (CoordFlag) {
      x1 = 1.0;
      x2 = (float) Nc;
      y1 = 1.0;
      y2 = (float) Nr;
      z1 = 1.0;
      z2 = (float) Nl;
    }
   else {
      x1 = WestLon[it];
      x2 = EastLon[it];
      y1 = NorthLat[it];
      y2 = SouthLat[it];
      z1 = BottomHgt[it];
      z2 = TopHgt[it];
   }

   float2string( x1, str );
   plot_string( str, Xmin-0.02, Ymin-0.1, Zmax+0.125, bx, ux, 0 );

   float2string( x2, str );
   plot_string( str, Xmax-0.05, Ymin-0.1, Zmax+0.125, bx, ux, 0 );

   float2string( y1, str );
   plot_string( str, Xmin-0.075, Ymax-0.03, Zmax+0.075, by, uy, 1 );

   float2string( y2, str );
   plot_string( str, Xmin-0.075, Ymin-0.02, Zmax+0.075, by, uy, 1 );

   float2string( z1, str );
   plot_string( str, Xmin-0.07, Ymin-0.07, Zmax+0.005, bz, uz, 1 );

   float2string( z2, str );
   plot_string( str, Xmin-0.07, Ymin-0.07, Zmin+0.005, bz, uz, 1 );

   set_depthcue( 0 );

}





/*** draw_cursor **********************************************************
   Draw the 3-D cursor.
   Input:  it - timestep
           labels - 0 = don't show cursor lables, non-zero = show labels
**************************************************************************/
static void draw_cursor( it, labels )
int it, labels;
{
   static float bx[3] = { 0.05, 0.0, 0.0 },      ux[3] = { 0.0, 0.05, -0.05 };
   static float by[3] = { -0.035, 0.0, 0.035 },  uy[3] = { 0.0, 0.07, 0.0 };
   static float bz[3] = { -0.035, -0.035, 0.0 }, uz[3] = { 0.0, 0.0, -0.07 };
   float v[6][3];
   float x, y, z;
   char str[100];

   set_color( *CursorColor );
   set_depthcue( 1 );

   /* draw cursor */
   if (RibbonFlag) {
      /* draw ribbon cursor */
      /* X-axis */
      v[0][0] = CursorX-0.05; v[0][1] = CursorY-0.005; v[0][2] = CursorZ+0.005;
      v[1][0] = CursorX-0.05; v[1][1] = CursorY+0.005; v[1][2] = CursorZ-0.005;
      v[2][0] = CursorX+0.05; v[2][1] = CursorY-0.005; v[2][2] = CursorZ+0.005;
      v[3][0] = CursorX+0.05; v[3][1] = CursorY+0.005; v[3][2] = CursorZ-0.005;
      polytri( v, 4 );
      v[0][0] = CursorX-0.05; v[0][1] = CursorY-0.005; v[0][2] = CursorZ-0.005;
      v[1][0] = CursorX-0.05; v[1][1] = CursorY+0.005; v[1][2] = CursorZ+0.005;
      v[2][0] = CursorX+0.05; v[2][1] = CursorY-0.005; v[2][2] = CursorZ-0.005;
      v[3][0] = CursorX+0.05; v[3][1] = CursorY+0.005; v[3][2] = CursorZ+0.005;
      polytri( v, 4 );

      /* Y-axis */
      v[0][0] = CursorX-0.005; v[0][1] = CursorY-0.05; v[0][2] = CursorZ+0.005;
      v[1][0] = CursorX+0.005; v[1][1] = CursorY-0.05; v[1][2] = CursorZ-0.005;
      v[2][0] = CursorX-0.005; v[2][1] = CursorY+0.05; v[2][2] = CursorZ+0.005;
      v[3][0] = CursorX+0.005; v[3][1] = CursorY+0.05; v[3][2] = CursorZ-0.005;
      polytri( v, 4 );
      v[0][0] = CursorX-0.005; v[0][1] = CursorY-0.05; v[0][2] = CursorZ-0.005;
      v[1][0] = CursorX+0.005; v[1][1] = CursorY-0.05; v[1][2] = CursorZ+0.005;
      v[2][0] = CursorX-0.005; v[2][1] = CursorY+0.05; v[2][2] = CursorZ-0.005;
      v[3][0] = CursorX+0.005; v[3][1] = CursorY+0.05; v[3][2] = CursorZ+0.005;
      polytri( v, 4 );

      /* Z-axis */
      v[0][0] = CursorX-0.005; v[0][1] = CursorY-0.005; v[0][2] = CursorZ+0.05;
      v[1][0] = CursorX+0.005; v[1][1] = CursorY+0.005; v[1][2] = CursorZ+0.05;
      v[2][0] = CursorX-0.005; v[2][1] = CursorY-0.005; v[2][2] = CursorZ-0.05;
      v[3][0] = CursorX+0.005; v[3][1] = CursorY+0.005; v[3][2] = CursorZ-0.05;
      polytri( v, 4 );
      v[0][0] = CursorX-0.005; v[0][1] = CursorY+0.005; v[0][2] = CursorZ+0.05;
      v[1][0] = CursorX+0.005; v[1][1] = CursorY-0.005; v[1][2] = CursorZ+0.05;
      v[2][0] = CursorX-0.005; v[2][1] = CursorY+0.005; v[2][2] = CursorZ-0.05;
      v[3][0] = CursorX+0.005; v[3][1] = CursorY-0.005; v[3][2] = CursorZ-0.05;
      polytri( v, 4 );

   }
   else {
      /* draw line segment cursor */
      v[0][0] = CursorX-0.05;   v[1][0] = CursorX+0.05;
      v[0][1] = v[1][1] = CursorY;
      v[0][2] = v[1][2] = CursorZ;

      v[2][0] = v[3][0] = CursorX;
      v[2][1] = CursorY-0.05;   v[3][1] = CursorY+0.05;
      v[2][2] = v[3][2] = CursorZ;

      v[4][0] = v[5][0] = CursorX;
      v[4][1] = v[5][1] = CursorY;
      v[4][2] = CursorZ-0.05;   v[5][2] = CursorZ+0.05;

      disjointpolyline( v, 6 );
   }

   if (labels) {
      /* draw tick marks */
      v[0][0] = v[1][0] = CursorX;
      v[0][1] = Ymin;  v[1][1] = Ymin-0.05;
      v[0][2] = Zmax;  v[1][2] = Zmax+0.05;
      v[2][0] = Xmin;  v[3][0] = Xmin-0.05;
      v[2][1] = v[3][1] = CursorY;
      v[2][2] = Zmax;  v[3][2] = Zmax+0.05;
      v[4][0] = Xmin;  v[5][0] = Xmin-0.05;
      v[4][1] = Ymin;  v[5][1] = Ymin-0.05;
      v[4][2] = v[5][2] = CursorZ;
      disjointpolyline( v, 6 );

      /* draw position labels */
      if (CoordFlag) {
	 x = CursorC+1.0;
	 y = CursorR+1.0;
	 z = CursorL+1.0;
      }
      else {
	 x = WestLon[it] - CursorC*LonInc;
	 y = NorthLat[it] - CursorR*LatInc;
	 z = BottomHgt[it] + CursorL*HgtInc;
      }

      float2string( x, str );
      plot_string( str, CursorX-0.04, Ymin-0.1, Zmax+0.125, bx, ux, 0 );
      float2string( y, str );
      plot_string( str, Xmin-0.075, CursorY-0.02, Zmax+0.075, by, uy, 1 );
      float2string( z, str );
      plot_string( str, Xmin-0.07, Ymin-0.07, CursorZ+0.005, bz, uz, 1 );
   }

   set_depthcue( 0 );

}



/* size of temporary arrays, MAXVERT, must be >= 2*MAX(Nr,Nc,Nl) */
/*#define MAXVERT 20000*/
#define MAXVERT 10000



/*** draw_surfaces ****************************************************
   Draw all activated 3-D contour surfaces.
   Input:  it - time step.
           tf - transparency flag: 1=only draw opaque surfaces
                                   0=only draw transparent surfaces
   Output:  the surfaces are drawn.
**********************************************************************/
static void draw_surfaces( it, tf )
int it, tf;
{
   int ip, i, j, jj, num, remaining;
   float vert[MAXVERT][3], norm[MAXVERT][3];
   float xm, ym, zm, nm;
   int_2 *vxptr, *vyptr, *vzptr;
   int_1 *nxptr, *nyptr, *nzptr;
   uint_2 *ptsptr;

   for (ip=0;ip<NumParms+NumClones;ip++) {
      if (DisplaySurf[ip] && SurfTable[ip][it].valid
	  && cond_lock(&SurfTable[ip][it].lock)) {

	 recent( ISOSURF, ip );

	 if ( (tf && (Color[ip][ISOSURF]&0xff000000)==0xff000000) ||
	      (!tf && (Color[ip][ISOSURF]&0xff000000)!=0xff000000) ) {
	    set_color( Color[ip][0] );

	    /* decompress vertices, normals and render surface */
	    xm = Gxscale / 10000.0;
	    ym = Gyscale / 10000.0;
	    zm = Gzscale / 10000.0;
	    nm = 1.0 / 125.0;

	    vxptr = SurfTable[ip][it].vx;
	    vyptr = SurfTable[ip][it].vy;
	    vzptr = SurfTable[ip][it].vz;
	    nxptr = SurfTable[ip][it].nx;
	    nyptr = SurfTable[ip][it].ny;
	    nzptr = SurfTable[ip][it].nz;
	    ptsptr = SurfTable[ip][it].pts;
	    remaining = SurfTable[ip][it].numpts;

	    j = 0;
	    while (remaining) {

	       num = MIN( remaining, MAXVERT );
               /*$dir no_recurrence*/
	       for (i=0;i<num;i++) {
		  jj = (int) ptsptr[j];
		  vert[i][0] = (float) vxptr[jj] * xm;
		  vert[i][1] = (float) vyptr[jj] * ym;
		  vert[i][2] = (float) vzptr[jj] * zm;
		  norm[i][0] = (float) nxptr[jj] * nm;
		  norm[i][1] = (float) nyptr[jj] * nm;
		  norm[i][2] = (float) nzptr[jj] * nm;
		  j++;
	       }
	       polytrinorm( vert, norm, num );

	       remaining -= num;
	       if (remaining) {
		  remaining += 2;
		  j -= 2;  /* duplicate last 2 vertices */
	       }

	    }

	 }

	 clear_lock( &SurfTable[ip][it].lock );
      }
   }
}




/*** draw_hslices *****************************************************
   Draw all activated horizontal contour slices.
   Input:  it - time step
           labels - draw labels flag.
   Output:  the slices are drawn.
**********************************************************************/
static void draw_hslices( it, labels )
int it, labels;
{
   int i, j, ip, remaining, num;
   int_2 *vxptr, *vyptr;
   float vert[MAXVERT][3], z, xm, ym;

   for (ip=0;ip<NumParms+NumClones;ip++) {
      if (DisplayHSlice[ip]) {
	 if (HSliceTable[ip][it].valid
	     && cond_lock(&HSliceTable[ip][it].lock)) {

	    recent( HSLICE, ip );
	    set_color( Color[ip][HSLICE] );

	    z = Zmax - HSliceTable[ip][it].level / (Nl-1) * (Zmax-Zmin);
	    xm = Gxscale / 10000.0;
	    ym = Gyscale / 10000.0;

	    /* draw contour lines */
	    vxptr = HSliceTable[ip][it].vx1;
	    vyptr = HSliceTable[ip][it].vy1;
	    remaining = HSliceTable[ip][it].num1;
	    j = 0;
	    while (remaining) {
	       num = MIN( remaining, MAXVERT );
	       /*$dir no_recurrence */
	       for (i=0;i<num;i++) {
		  vert[i][0] = (float) vxptr[j] * xm;
		  vert[i][1] = (float) vyptr[j] * ym;
		  vert[i][2] = z;
		  j++;
	       }
	       disjointpolyline( vert, num );
	       remaining -= num;
	    }
	    
	    if (labels) {
	       /* draw contour labels */
	       vxptr = HSliceTable[ip][it].vx3;
	       vyptr = HSliceTable[ip][it].vy3;
	       remaining = HSliceTable[ip][it].num3;
	       j = 0;
	       while (remaining) {
		  num = MIN( remaining, MAXVERT );
		  /*$dir no_recurrence */
		  for (i=0;i<num;i++) {
		     vert[i][0] = (float) vxptr[j] * xm;
		     vert[i][1] = (float) vyptr[j] * ym;
		     vert[i][2] = z;
		     j++;
		  }
		  disjointpolyline( vert, num );
		  remaining -= num;
	       }
	    }
	    else {
	       /* draw hidden contour lines */
	       vxptr = HSliceTable[ip][it].vx2;
	       vyptr = HSliceTable[ip][it].vy2;
	       remaining = HSliceTable[ip][it].num2;
	       j = 0;
	       while (remaining) {
		  num = MIN( remaining, MAXVERT );
		  /*$dir no_recurrence */
		  for (i=0;i<num;i++) {
		     vert[i][0] = (float) vxptr[j] * xm;
		     vert[i][1] = (float) vyptr[j] * ym;
		     vert[i][2] = z;
		     j++;
		  }
		  disjointpolyline( vert, num );
		  remaining -= num;
	       }
	    }
	    clear_lock( &HSliceTable[ip][it].lock );
	 }

	 /* draw position label */
	 draw_slice_tick( 'z', HSliceLevel[ip], Color[ip][HSLICE], it );

      }
   }

}






/*** draw_vslices *****************************************************
   Draw all activated vertical contour slices.
   Input:  it - time step.
           labels - draw labels flag.
   Output:  the slices are drawn.
**********************************************************************/
static void draw_vslices( it, labels )
int it, labels;
{
   int i, j, ip, num, remaining;
   int_2 *xptr, *yptr, *zptr;
   float vert[MAXVERT][3], xm, ym, zm;

   for (ip=0;ip<NumParms+NumClones;ip++) {
      if (DisplayVSlice[ip]) {
	 if (VSliceTable[ip][it].valid
	     && cond_lock(&VSliceTable[ip][it].lock)) {

	    recent( VSLICE, ip );
	    set_color( Color[ip][VSLICE] );

	    xm = Gxscale / 10000.0;
	    ym = Gyscale / 10000.0;
	    zm = Gzscale / 10000.0;

	    /* draw contour lines */
	    xptr = VSliceTable[ip][it].vx1;
	    yptr = VSliceTable[ip][it].vy1;
	    zptr = VSliceTable[ip][it].vz1;
	    remaining = VSliceTable[ip][it].num1;
	    j = 0;
	    while (remaining) {
	       num = MIN( remaining, MAXVERT );
	       /*$dir no_recurrence */
	       for (i=0;i<num;i++) {
		  vert[i][0] = (float) xptr[j] * xm;
		  vert[i][1] = (float) yptr[j] * ym;
		  vert[i][2] = (float) zptr[j] * zm;
		  j++;
	       }
	       disjointpolyline( vert, num );
	       remaining -= num;
	    }

	    if (labels) {
	       /* draw contour labels */
	       xptr = VSliceTable[ip][it].vx3;
	       yptr = VSliceTable[ip][it].vy3;
	       zptr = VSliceTable[ip][it].vz3;
	       remaining = VSliceTable[ip][it].num3;
	       j = 0;
	       while (remaining) {
		  num = MIN( remaining, MAXVERT );
		  /*$dir no_recurrence */
		  for (i=0;i<num;i++) {
		     vert[i][0] = (float) xptr[j] * xm;
		     vert[i][1] = (float) yptr[j] * ym;
		     vert[i][2] = (float) zptr[j] * zm;
		     j++;
		  }
		  disjointpolyline( vert, num );
		  remaining -= num;
	       }
	    }
	    else {
	       /* draw hidden contour lines */
	       xptr = VSliceTable[ip][it].vx2;
	       yptr = VSliceTable[ip][it].vy2;
	       zptr = VSliceTable[ip][it].vz2;
	       remaining = VSliceTable[ip][it].num2;
	       j = 0;
	       while (remaining) {
		  num = MIN( remaining, MAXVERT );
		  /*$dir no_recurrence */
		  for (i=0;i<num;i++) {
		     vert[i][0] = (float) xptr[j] * xm;
		     vert[i][1] = (float) yptr[j] * ym;
		     vert[i][2] = (float) zptr[j] * zm;
		     j++;
		  }
		  disjointpolyline( vert, num );
		  remaining -= num;
	       }
	    }
	    clear_lock( &VSliceTable[ip][it].lock );
	 }

	 /* draw position labels */
	 if (VSliceC1[ip]==0.0) {
	    draw_slice_tick( 'W', VSliceR1[ip], Color[ip][VSLICE], it );
	 }
	 else if (VSliceC1[ip]==(float)(Nc-1)) {
	    draw_slice_tick( 'E', VSliceR1[ip], Color[ip][VSLICE], it );
	 }
	 else if (VSliceR1[ip]==0.0) {
	    draw_slice_tick( 'N', VSliceC1[ip], Color[ip][VSLICE], it );
	 }
	 else if (VSliceR1[ip]==(float)(Nr-1)) {
	    draw_slice_tick( 'S', VSliceC1[ip], Color[ip][VSLICE], it );
	 }
	 if (VSliceC2[ip]==0.0) {
	    draw_slice_tick( 'W', VSliceR2[ip], Color[ip][VSLICE], it );
	 }
	 else if (VSliceC2[ip]==(float)(Nc-1)) {
	    draw_slice_tick( 'E', VSliceR2[ip], Color[ip][VSLICE], it );
	 }
	 else if (VSliceR2[ip]==0.0) {
	    draw_slice_tick( 'N', VSliceC2[ip], Color[ip][VSLICE], it );
	 }
	 else if (VSliceR2[ip]==(float)(Nr-1)) {
	    draw_slice_tick( 'S', VSliceC2[ip], Color[ip][VSLICE], it );
	 }

      }
   }

}





/*** draw_chslices ****************************************************
   Draw all activated horizontal color slices.
   Input:  it - time step.
           tf - transparency flag: 1=only draw opaque slices
                                   0=only draw transparent slices
   Output:  the slices are drawn.
**********************************************************************/
static void draw_chslices( it, tf )
int it, tf;
{
   uint_1 *data;
   int i, j, ip, ir, ic, rows;
   int color[MAXVERT];
   float vert[MAXVERT][3], x, y, z, dx, dy;

   for (ip=0;ip<NumParms+NumClones;ip++) {
      if (DisplayCHSlice[ip]) {
	 if (CHSliceTable[ip][it].valid
	     && cond_lock(&CHSliceTable[ip][it].lock)) {

	    recent( CHSLICE, ip );
	    if ( (tf && (Color[ip][CHSLICE]&0xff000000)==0xff000000) ||
		(!tf && (Color[ip][CHSLICE]&0xff000000)!=0xff000000) ) {

	       set_color( Color[ip][CHSLICE] );

	       z = Zmax - CHSliceTable[ip][it].level / (Nl-1) * (Zmax-Zmin);

	       data = CHSliceTable[ip][it].data;

	       /* Build a quad-mesh */
	       dx = (Xmax-Xmin) / (Nc-1);
	       dy = (Ymin-Ymax) / (Nr-1);
	       i = j = 0;
	       rows = 0;
	       y = Ymax;
	       for (ir=0;ir<Nr;ir++) {
		  /* do a row */
		  x = Xmin;
		  /*$dir scalar */
		  for (ic=0;ic<Nc;ic++) {
		     vert[i][0] = x;
		     vert[i][1] = y;
		     vert[i][2] = z;
		     color[i] = CHSliceColors[ip][ (int) data[j] ];
		     x += dx;
		     i++;
		     j++;
		  }

		  /* check for array overflow */
		  if (i+Nc>MAXVERT) {
		     /* another iteration will cause overflow */
		     quadmesh( vert, color, rows+1, Nc );
		     rows = 0;
		     ir--;
		     i = 0;
		     j -= Nc;
		  }
		  else {
		     /* continue filling arrays */
		     y += dy;
		     rows++;
		  }
	       }

	       quadmesh( vert, color, rows, Nc );

	    }
	    clear_lock( &CHSliceTable[ip][it].lock );
	 }
	 if (tf) {
	    draw_slice_tick( 'z', CHSliceLevel[ip], Color[ip][CHSLICE], it );
	 }
      }
   }

}




/*** draw_cvslices ****************************************************
   Draw all activated vertical color slices.
   Input:  it - time step.
           tf - transparency flag: 1=only draw opaque slices
                                   0=only draw transparent slices
   Output:  the slices are drawn.
**********************************************************************/
static void draw_cvslices( it, tf )
int it, tf;
{
   uint_1 *data;
   int i, j, ip, ir, ic, rows, cols, count;
   int color[MAXVERT];
   float vert[MAXVERT][3], x,y,z, x1,y1,z1, x2,y2,z2, dx,dy,dz;

   for (ip=0;ip<NumParms+NumClones;ip++) {
      if (DisplayCVSlice[ip]) {
	 if (CVSliceTable[ip][it].valid
	     && cond_lock(&CVSliceTable[ip][it].lock)) {

	    recent( CVSLICE, ip );

	    if ( (tf && (Color[ip][CVSLICE]&0xff000000)==0xff000000) ||
		(!tf && (Color[ip][CVSLICE]&0xff000000)!=0xff000000) ) {

	       set_color( Color[ip][CVSLICE] );

	       rows = CVSliceTable[ip][it].rows;
	       cols = CVSliceTable[ip][it].cols;
	       x1 = CVSliceTable[ip][it].c1 / (Nc-1) * (Xmax-Xmin) + Xmin;
	       y1 = CVSliceTable[ip][it].r1 / (Nr-1) * (Ymin-Ymax) + Ymax;
	       z1 = Zmax;
	       x2 = CVSliceTable[ip][it].c2 / (Nc-1) * (Xmax-Xmin) + Xmin;
	       y2 = CVSliceTable[ip][it].r2 / (Nr-1) * (Ymin-Ymax) + Ymax;
	       z2 = Zmin;
	       dx = (x2-x1) / (cols-1);
	       dy = (y2-y1) / (cols-1);
	       dz = (z2-z1) / (rows-1);

	       data = CVSliceTable[ip][it].data;

	       /* Build a quad-mesh */
	       i = j = 0;
	       count = 0;
	       z = z1;
	       for (ir=0;ir<rows;ir++) {
		  /* do a row */
		  x = x1;
		  y = y1;
		  /*$dir scalar */
		  for (ic=0;ic<cols;ic++) {
		     vert[i][0] = x;
		     vert[i][1] = y;
		     vert[i][2] = z;
		     color[i] = CVSliceColors[ip][ (int) data[j] ];
		     x += dx;
		     y += dy;
		     i++;
		     j++;
		  }

		  /* check for array overflow */
		  if (i+cols>MAXVERT) {
		     /* another iteration will cause overflow */
		     quadmesh( vert, color, count+1, cols );
		     count = 0;
		     ir--;
		     i = 0;
		     j -= cols;
		  }
		  else {
		     /* continue filling arrays */
		     x += dx;
		     y += dy;
		     count++;
		  }
		  z += dz;
	       }

	       quadmesh( vert, color, count, cols );

	    }
	    clear_lock( &CVSliceTable[ip][it].lock );

	 }

	 if (tf) {
	    /* draw position labels */
	    if (CVSliceC1[ip]==0.0) {
	       draw_slice_tick( 'W', CVSliceR1[ip], Color[ip][CVSLICE], it );
	    }
	    else if (CVSliceC1[ip]==(float)(Nc-1)) {
	       draw_slice_tick( 'E', CVSliceR1[ip], Color[ip][CVSLICE], it );
	    }
	    else if (CVSliceR1[ip]==0.0) {
	       draw_slice_tick( 'N', CVSliceC1[ip], Color[ip][CVSLICE], it );
	    }
	    else if (CVSliceR1[ip]==(float)(Nr-1)) {
	       draw_slice_tick( 'S', CVSliceC1[ip], Color[ip][CVSLICE], it );
	    }
	    if (CVSliceC2[ip]==0.0) {
	       draw_slice_tick( 'W', CVSliceR2[ip], Color[ip][CVSLICE], it );
	    }
	    else if (CVSliceC2[ip]==(float)(Nc-1)) {
	       draw_slice_tick( 'E', CVSliceR2[ip], Color[ip][CVSLICE], it );
	    }
	    else if (CVSliceR2[ip]==0.0) {
	       draw_slice_tick( 'N', CVSliceC2[ip], Color[ip][CVSLICE], it );
	    }
	    else if (CVSliceR2[ip]==(float)(Nr-1)) {
	       draw_slice_tick( 'S', CVSliceC2[ip], Color[ip][CVSLICE], it );
	    }
	 }

      }
   }

}



static void draw_hwind( it )
int it;
{
   int i, j, ip, num, remaining, w;
   int_2 *vxptr, *vyptr, *vzptr;
   float vert[MAXVERT][3], xm, ym, zm;
   float z;

   for (w=0;w<WINDSLICES;w++) {

      if (DisplayHWind[w]) {
	 if (HWindTable[w][it].valid && cond_lock(&HWindTable[w][it].lock)) {

	    recent( HWIND, w );
	    set_color( HWindColor[w] );

	    /* draw bounding rectangle */
	    z = Zmax - HWindTable[w][it].level / (Nl-1) * (Zmax-Zmin);
	    vert[0][0] = Xmax;  vert[0][1] = Ymax;  vert[0][2] = z;
	    vert[1][0] = Xmax;  vert[1][1] = Ymin;  vert[1][2] = z;
	    vert[2][0] = Xmin;  vert[2][1] = Ymin;  vert[2][2] = z;
	    vert[3][0] = Xmin;  vert[3][1] = Ymax;  vert[3][2] = z;
	    vert[4][0] = Xmax;  vert[4][1] = Ymax;  vert[4][2] = z;
	    polyline( vert, 5 );

	    xm = Gxscale / 10000.0;
	    ym = Gyscale / 10000.0;
	    zm = Gzscale / 10000.0;

	    /* draw wind vectors */
	    vxptr = HWindTable[w][it].vx;
	    vyptr = HWindTable[w][it].vy;
	    vzptr = HWindTable[w][it].vz;
	    remaining = HWindTable[w][it].nvectors;
	    j = 0;
	    while (remaining>0) {
	       num = MIN( remaining, MAXVERT/6 );  /*num = number of vectors*/
	       /*$dir no_recurrence */
	       for (i=0;i<num*6;) {
		  /* main vector */
		  vert[i][0] = (float) vxptr[j] * xm;
		  vert[i][1] = (float) vyptr[j] * ym;
		  vert[i][2] = (float) vzptr[j] * zm;
		  i++;
		  vert[i][0] = (float) vxptr[j+1] * xm;
		  vert[i][1] = (float) vyptr[j+1] * ym;
		  vert[i][2] = (float) vzptr[j+1] * zm;
		  i++;
		  /* left half of arrow head */
		  vert[i][0] = (float) vxptr[j+1] * xm;
		  vert[i][1] = (float) vyptr[j+1] * ym;
		  vert[i][2] = (float) vzptr[j+1] * zm;
		  i++;
		  vert[i][0] = (float) vxptr[j+2] * xm;
		  vert[i][1] = (float) vyptr[j+2] * ym;
		  vert[i][2] = (float) vzptr[j+2] * zm;
		  i++;
		  /* right half of arrow head */
		  vert[i][0] = (float) vxptr[j+1] * xm;
		  vert[i][1] = (float) vyptr[j+1] * ym;
		  vert[i][2] = (float) vzptr[j+1] * zm;
		  i++;
		  vert[i][0] = (float) vxptr[j+3] * xm;
		  vert[i][1] = (float) vyptr[j+3] * ym;
		  vert[i][2] = (float) vzptr[j+3] * zm;
		  i++;
		  j += 4;
	       }
	       disjointpolyline( vert, num*6 );
	       remaining -= num;
	    }
	    clear_lock( &HWindTable[w][it].lock );
	 }

	 /* draw position label */
	 draw_slice_tick( 'z', HWindLevel[w], HWindColor[w], it );

      }

   }
}



static void draw_vwind( it )
int it;
{
   int i, j, ip, num, remaining, w;
   int_2 *vxptr, *vyptr, *vzptr;
   float vert[MAXVERT][3], xm, ym, zm;
   float x1,y1, x2,y2;

   for (w=0;w<WINDSLICES;w++) {
      if (DisplayVWind[w]) {
	 if (VWindTable[w][it].valid && cond_lock(&VWindTable[w][it].lock)) {

	    recent( VWIND, w );
	    set_color( VWindColor[w] );

	    /* draw bounding rectangle */
	    x1 = Xmin + VWindTable[w][it].c1 / (Nc-1) * (Xmax-Xmin);
	    x2 = Xmin + VWindTable[w][it].c2 / (Nc-1) * (Xmax-Xmin);
	    y1 = Ymax - VWindTable[w][it].r1 / (Nr-1) * (Ymax-Ymin);
	    y2 = Ymax - VWindTable[w][it].r2 / (Nr-1) * (Ymax-Ymin);
	    vert[0][0] = x1;  vert[0][1] = y1;  vert[0][2] = Zmin;
	    vert[1][0] = x2;  vert[1][1] = y2;  vert[1][2] = Zmin;
	    vert[2][0] = x2;  vert[2][1] = y2;  vert[2][2] = Zmax;
	    vert[3][0] = x1;  vert[3][1] = y1;  vert[3][2] = Zmax;
	    vert[4][0] = x1;  vert[4][1] = y1;  vert[4][2] = Zmin;
	    polyline( vert, 5 );

	    xm = Gxscale / 10000.0;
	    ym = Gyscale / 10000.0;
	    zm = Gzscale / 10000.0;

	    /* draw wind vector lines */
	    vxptr = VWindTable[w][it].vx;
	    vyptr = VWindTable[w][it].vy;
	    vzptr = VWindTable[w][it].vz;
	    remaining = VWindTable[w][it].nvectors;
	    j = 0;
	    while (remaining>0) {
	       num = MIN( remaining, MAXVERT/6 );  /*num = number of vectors*/
	       /*$dir no_recurrence */
	       for (i=0;i<num*6;) {
		  /* main vector */
		  vert[i][0] = (float) vxptr[j] * xm;
		  vert[i][1] = (float) vyptr[j] * ym;
		  vert[i][2] = (float) vzptr[j] * zm;
		  i++;
		  vert[i][0] = (float) vxptr[j+1] * xm;
		  vert[i][1] = (float) vyptr[j+1] * ym;
		  vert[i][2] = (float) vzptr[j+1] * zm;
		  i++;
		  /* left half of arrow head */
		  vert[i][0] = (float) vxptr[j+1] * xm;
		  vert[i][1] = (float) vyptr[j+1] * ym;
		  vert[i][2] = (float) vzptr[j+1] * zm;
		  i++;
		  vert[i][0] = (float) vxptr[j+2] * xm;
		  vert[i][1] = (float) vyptr[j+2] * ym;
		  vert[i][2] = (float) vzptr[j+2] * zm;
		  i++;
		  /* right half of arrow head */
		  vert[i][0] = (float) vxptr[j+1] * xm;
		  vert[i][1] = (float) vyptr[j+1] * ym;
		  vert[i][2] = (float) vzptr[j+1] * zm;
		  i++;
		  vert[i][0] = (float) vxptr[j+3] * xm;
		  vert[i][1] = (float) vyptr[j+3] * ym;
		  vert[i][2] = (float) vzptr[j+3] * zm;
		  i++;
		  j += 4;
	       }
	       disjointpolyline( vert, num*6 );
	       remaining -= num;
	    }
	    clear_lock( &VWindTable[w][it].lock );

	 }

	 /* position labels */
	 if (VWindC1[w]==0.0) {
	    draw_slice_tick( 'W', VWindR1[w], VWindColor[w], it );
	 }
	 else if (VWindC1[w]==(float)(Nc-1)) {
	    draw_slice_tick( 'E', VWindR1[w], VWindColor[w], it );
	 }
	 else if (VWindR1[w]==0.0) {
	    draw_slice_tick( 'N', VWindC1[w], VWindColor[w], it );
	 }
	 else if (VWindR1[w]==(float)(Nr-1)) {
	    draw_slice_tick( 'S', VWindC1[w], VWindColor[w], it );
	 }
	 if (VWindC2[w]==0.0) {
	    draw_slice_tick( 'W', VWindR2[w], VWindColor[w], it );
	 }
	 else if (VWindC2[w]==(float)(Nc-1)) {
	    draw_slice_tick( 'E', VWindR2[w], VWindColor[w], it );
	 }
	 else if (VWindR2[w]==0.0) {
	    draw_slice_tick( 'N', VWindC2[w], VWindColor[w], it );
	 }
	 else if (VWindR2[w]==(float)(Nr-1)) {
	    draw_slice_tick( 'S', VWindC2[w], VWindColor[w], it );
	 }

      }
   }

}



static void draw_traj( it )
int it;
{
   int i, j, k, len, start;
   float vert[MAXVERT][3], norm[MAXVERT][3];
   float xm, ym, zm, nm;
   int_2 *vxptr, *vyptr, *vzptr;
   int_1 *nxptr, *nyptr, *nzptr;

   xm = Gxscale / 10000.0;
   ym = Gyscale / 10000.0;
   zm = Gzscale / 10000.0;
   nm = 1.0 / 125.0;
   
   for (i=0;i<NumTraj;i++) {

      if (DisplayTraj[TrajTable[i].group]
	  && cond_lock(&TrajTable[i].lock) ) {
	 assert( TrajTable[i].lock==1 );

	 recent( TRAJ, TrajTable[i].group );
	 set_color( TrajColor[TrajTable[i].group] );

	 vxptr = TrajTable[i].vx;
	 vyptr = TrajTable[i].vy;
	 vzptr = TrajTable[i].vz;
	 nxptr = TrajTable[i].nx;
	 nyptr = TrajTable[i].ny;
	 nzptr = TrajTable[i].nz;
	 start = TrajTable[i].start[it];
	 len = MIN( TrajTable[i].len[it], MAXVERT );

	 if (start!=0xffff) {
	    j = start;
	    /*$dir no_recurrence*/
	    for (k=0; k<len; k++) {
	       vert[k][0] = (float) vxptr[j] * xm;
	       vert[k][1] = (float) vyptr[j] * ym;
	       vert[k][2] = (float) vzptr[j] * zm;
	       j++;
	    }
	    if (TrajTable[i].kind==0) {
	       /* trajectory lines */
	       if (Debug) {
		  printf("len=%d\n",len);
		  for (k=0;k<len;k++) {
		   printf("%d: %f %f %f\n",k,vert[k][0],vert[k][1],vert[k][2]);
		  }
	       }
	       polyline( vert, len );
	    }
	    else {
	       /* trajectory ribbon -> uncompress normals */
	       j = start;
	       /*$dir no_recurrence*/
	       for (k=0;k<len;k++) {
		  norm[k][0] = (float) nxptr[j] * nm;
		  norm[k][1] = (float) nyptr[j] * nm;
		  norm[k][2] = (float) nzptr[j] * nm;
		  j++;
	       }
	       polytrinorm( vert, norm, len );
	    }
	 }

	 clear_lock( &TrajTable[i].lock );
      }

   }

}




/*** draw_clock *******************************************************
   Draw the clock in the upper-left corner of the 3-D window.
   Input:  c - the color to use.
   Output:  the clock is drawn.
**********************************************************************/
static void draw_clock( c )
unsigned long c;
{
   static float twopi = 2.0 * 3.141592;
   short pp[8][2];
   float ang;
   char str[12];
   int i;

   /* Draw the clock. */
   if (NumTimes)
      ang = (float) CurTime / (float) NumTimes;
   else
      ang = 0.0;

   pp[0][1] = 50;
   pp[0][0] = 50;
   pp[1][1] = 50 - 40 * cos(twopi * ang);
   pp[1][0] = 50 + 40 * sin(twopi * ang);
   pp[2][1] = pp[1][1] + 1;
   pp[2][0] = pp[1][0] + 1;
   pp[3][1] = pp[0][1] + 1;
   pp[3][0] = pp[0][0] + 1;
   pp[4][1] = pp[0][1] - 1;
   pp[4][0] = pp[0][0] + 1;
   pp[5][1] = pp[1][1] - 1;
   pp[5][0] = pp[1][0] + 1;

   set_color( c );
   polyline2d( pp, 6 );

   i = TimeStamp[CurTime];
   sprintf( str, "%02d:%02d:%02d", i/3600, (i/60)%60, i%60 );
   draw_text( 100, FontHeight+5, str );

   sprintf( str, "%05d", yyddd( DayStamp[CurTime] ) );
   draw_text( 100, 2*FontHeight+10, str );

   sprintf( str, "%d / %d", CurTime+1, NumTimes );
   draw_text( 100, 3*FontHeight+15, str );

}



/*** draw_info ********************************************************
   Render status info at bottom of window.
**********************************************************************/
static void draw_info()
{
   char str[1000];
   int m;

   m = mem_used();
   if (m>=0)
      sprintf(str, "Pending: %d   Memory Used: %d", get_qsize(), m );
   else
      sprintf(str, "Pending: %d", get_qsize() );

   draw_text( 10, Height - FontHeight, str );
}




draw_logo( c )
unsigned long c;
{
   static XPoint vv[7] = {{-1,0}, {9, 41}, {11, 41}, {21, 0},
                          {20, 0}, {10, 41}, {0, 0}};
   static XPoint ii[12] = {{0,0}, {21, 0}, {21, 1}, {11, 1},
                           {11, 40}, {21, 40}, {21, 41}, {0, 41},
                           {0, 40}, {10, 40}, {10, 1}, {0, 1}};
   static XPoint ss[24] = {{0,35}, {5, 40}, {16,40}, {20, 36},
                           {20, 25}, {16, 21}, {5, 21}, {0,16},
                           {0, 5}, {5, 0}, {15, 0}, {20, 5},
                           {20, 6}, {15, 1}, {5, 1}, {1, 5},
                           {1, 16}, {5, 20}, {16, 20}, {21, 25},
                           {21, 36}, {16, 41}, {5, 41}, {0, 36}};
   static XPoint sh[4] = {{0, 20}, {16, 20}, {16, 21}, {0, 21}};
   static XPoint s5[18] = {{0,35}, {5, 40}, {16,40}, {20, 36},
                           {20, 25}, {16, 21}, {0, 21},
                           {0, 0}, {20, 0}, {20,1}, {1, 1},
                           {1, 20}, {16, 20}, {21, 25},
                           {21, 36}, {16, 41}, {5, 41}, {0, 36}};
   static XPoint dd[14] = {{0, 0}, {0, 41}, {16, 41}, {21, 36},
                           {21, 5}, {16, 0}, {1, 0}, {1, 1},
                           {16, 1}, {20, 5}, {20, 36}, {16, 40},
                           {1, 40}, {1, 1}};
   short p[30][2];
   int j;

   set_color( c );

   for (j=0; j<7; j++) {
      p[j][0] = Width - 180 + vv[j].x;
      p[j][1] = Height - 50 + vv[j].y;
   }
   polyline2d( p, 7 );

   for (j=0; j<12; j++) {
      p[j][0] = Width - 150 + ii[j].x;
      p[j][1] = Height - 50 + ii[j].y;
   }
   polyline2d( p, 12 );

   for (j=0; j<24; j++) {
      p[j][0] = Width - 120 + ss[j].x;
      p[j][1] = Height - 50 + ss[j].y;
   }
   polyline2d( p, 24 );

   for (j=0; j<4; j++) {
      p[j][0] = Width - 90 + sh[j].x;
      p[j][1] = Height - 50 + sh[j].y;
   }
   polyline2d( p, 4 );

   for (j=0; j<18; j++) {
      p[j][0] = Width - 60 + s5[j].x;
      p[j][1] = Height - 50 + s5[j].y;
   }
   polyline2d( p, 18 );

   for (j=0; j<14; j++) {
      p[j][0] = Width - 30 + dd[j].x;
      p[j][1] = Height - 50 + dd[j].y;
   }
   polyline2d( p, 14 );

}




/*** draw_slice_tick **************************************************
   Draw a slice's tick mark.
   Input:  axis - 'W', 'E', 'S', 'N', or 'H'.
           pos - position in grid coordinates.
           color - drawing color
**********************************************************************/
static void draw_slice_tick( axis, pos, color, it )
char axis;
float pos;
unsigned long color;
int it;
{
   float x, y, z, level, tick[2][3];
   char str[100];
   /* base line and up vectors for vector text drawing */
   static float b1[3] = { 0.035, -0.035, 0.0 }, u1[3] = { 0.0, 0.0, -0.07 };
   static float b2[3] = { 0.05, 0.0, 0.0 }, u2[3] = { 0.0, 0.05, 0.0 };
   static float b3[3] = { -0.05, 0.0, 0.0 }, u3[3] = { 0.0, 0.05, 0.0 };

   set_color( color );

   if (axis=='W') {
      /* Top-West edge */
      y = pos / (Nr-1) * (Ymin-Ymax) + Ymax;
      tick[0][0] = Xmin;
      tick[0][1] = y;
      tick[0][2] = Zmin;
      tick[1][0] = Xmin-0.05;
      tick[1][1] = y;
      tick[1][2] = Zmin;
      polyline( tick, 2 );
      if (CoordFlag) {
	 level = pos+1.0;
      }
      else {
	 level = (pos / (Nr-1)) * (SouthLat[it]-NorthLat[it]) + NorthLat[it];
      }
      float2string( level, str );
      plot_string( str, Xmin-0.07, y, Zmin, b3,u3, 1 );
   }
   else if (axis=='E') {
      /* Top-East edge */
      y = pos / (Nr-1) * (Ymin-Ymax) + Ymax;
      tick[0][0] = Xmax;
      tick[0][1] = y;
      tick[0][2] = Zmin;
      tick[1][0] = Xmax+0.05;
      tick[1][1] = y;
      tick[1][2] = Zmin;
      polyline( tick, 2 );
      if (CoordFlag) {
	 level = pos+1.0;
      }
      else {
	 level = (pos / (Nr-1)) * (SouthLat[it]-NorthLat[it]) + NorthLat[it];
      }
      float2string( level, str );
      plot_string( str, Xmax+0.07, y, Zmin, b2,u2, 0 );
   }
   else if (axis=='S') {
      /* Top-South edge */
      x = pos / (Nc-1) * (Xmax-Xmin) + Xmin;
      tick[0][0] = x;
      tick[0][1] = Ymin;
      tick[0][2] = Zmin;
      tick[1][0] = x;
      tick[1][1] = Ymin-0.05;
      tick[1][2] = Zmin;
      polyline( tick, 2 );
      if (CoordFlag) {
	 level = pos+1.0;
      }
      else {
	 level = (pos / (Nc-1)) * (EastLon[it]-WestLon[it]) + WestLon[it];
      }
      float2string( level, str );
      plot_string( str, x-0.07, Ymin-0.12, Zmin, b2,u2, 0 );
   }
   else if (axis=='N') {
      /* Top-North edge */
      x = pos / (Nc-1) * (Xmax-Xmin) + Xmin;
      tick[0][0] = x;
      tick[0][1] = Ymax;
      tick[0][2] = Zmin;
      tick[1][0] = x;
      tick[1][1] = Ymax+0.05;
      tick[1][2] = Zmin;
      polyline( tick, 2 );
      if (CoordFlag) {
	 level = pos+1.0;
      }
      else {
	 level = (pos / (Nc-1)) * (EastLon[it]-WestLon[it]) + WestLon[it];
      }
      float2string( level, str );
      plot_string( str, x-0.07, Ymax+0.07, Zmin, b2,u2, 0 );
   }
   else if (axis=='z') {
      /* South-East edge */
      z = Zmax - pos / (Nl-1) * (Zmax-Zmin);
      tick[0][0] = Xmax;
      tick[0][1] = Ymin;
      tick[0][2] = z;
      tick[1][0] = Xmax + 0.05;
      tick[1][1] = Ymin - 0.05;
      tick[1][2] = z;
      polyline( tick, 2 );
      if (CoordFlag) {
	 level = pos+1.0;
      }
      else {
	 level = (pos / (Nl-1)) * (TopHgt[it]-BottomHgt[it]) + BottomHgt[it];
      }
      float2string( level, str );
      plot_string( str, Xmax+0.07, Ymin-0.07, z, b1,u1, 0 );
   }

}




/*** plot_string  ******************************************************
   Plot (draw) a string in 3-D.  At this time, only strings of digits,
   periods, and dashes are implemented.
   Input: f - the string to plot.
          startx, y, z - the point in 3-D to start at.
          base - vector indicating text baseline.
          up - vector indicating upright direction for text.
          rjustify - non-zero value indicates right justify the text.
**********************************************************************/
void plot_string( str, startx, starty, startz, base, up, rjustify )
char str[];
float startx, starty, startz;
float base[], up[];
int rjustify;
{
   static float zero[5][2] = { 0,0, 0,.8, .4,.8, .4,0, 0,0 },
      one[2][2] = { 0,0, 0,.8 },
      two[6][2] = { .4,0, 0,0, 0,.4, .4,.4, .4,.8, 0,.8 },
      three[7][2] = { 0,0, .4,0, .4,.4, 0,.4, .4,.4, .4,.8, 0,.8 },
      four[5][2] = { 0,.8, 0,.4, .4,.4, .4,.8, .4,0 },
      five[6][2] = { 0,0, .4,0, .4,.4, 0,.4, 0,.8, .4,.8 },
      six[6][2] = { .4,.8, 0,.8, 0,0, .4,0, .4,.4, 0,.4 },
      seven[4][2] = { 0,.7, 0,.8, .4,.8, .4,0 },
      eight[7][2] = { 0,0, 0,.8, .4,.8, .4,0, 0,0, 0,.4, .4,.4 },
      nine[5][2] = { .4,.4, 0,.4, 0,.8, .4,.8, .4,0 },
      dash[2][2] = { 0,.4, .4,.4 },
      dot[5][2] = { 0,0, 0,.1, .1,.1, .1,0, 0,0 };

   static float *index[12] = { (float*)zero, (float*)one, (float*)two,
      (float*)three, (float*)four, (float*)five, (float*)six, (float*)seven,
      (float*)eight, (float*)nine, (float*)dash, (float*)dot };

   static float width[12] = { 0.6, 0.2, 0.6, 0.6, 0.6, 0.6, 0.6, 0.6,
      0.6, 0.6, 0.6, 0.3 };
   static int verts[12] = { 5, 2, 6, 7, 5, 6, 6, 4, 7, 5, 2, 5 };

   float (*temp)[][2], plot[100][3];
   float cx, cy, cz;
   int i, j, k, len;

   cx = startx;  cy = starty;  cz = startz;
   len = strlen(str);

   if (rjustify) {
      /* draw right justified text */
      for (i=len-1; i>=0; i--) {
         if (str[i]=='-')
            k = 10;
         else if (str[i]=='.')
            k = 11;
         else if (str[i]>='0' && str[i]<='9')
            k = str[i] - '0';
         else
            continue;
         /* calculate position for this char */
         cx += width[k]*base[0];
         cy += width[k]*base[1];
         cz += width[k]*base[2];
         /* make the vertex array for this character */
         temp = (float***)index[k];
         for (j=0; j<verts[k]; j++) {
            plot[j][0] = cx - (*temp)[j][0]*base[0] + (*temp)[j][1]*up[0];
            plot[j][1] = cy - (*temp)[j][0]*base[1] + (*temp)[j][1]*up[1];
            plot[j][2] = cz - (*temp)[j][0]*base[2] + (*temp)[j][1]*up[2];
         }
	 polyline( plot, verts[k] );
      }

   }
   else {
      /* draw left justified text */
      for (i=0; i<len; i++) {
         if (str[i]=='-')
            k = 10;
         else if (str[i]=='.')
            k = 11;
         else if (str[i]>='0' && str[i]<='9')
            k = str[i] - '0';
         else
            continue;
         /* make the vertex array for this character */
         temp = (float***)index[k];
         for (j=0; j<verts[k]; j++) {
            plot[j][0] = cx + (*temp)[j][0]*base[0] + (*temp)[j][1]*up[0];
            plot[j][1] = cy + (*temp)[j][0]*base[1] + (*temp)[j][1]*up[1];
            plot[j][2] = cz + (*temp)[j][0]*base[2] + (*temp)[j][1]*up[2];
         }
	 polyline( plot, verts[k] );
         /* calculate position for next char */
         cx += width[k]*base[0];
         cy += width[k]*base[1];
         cz += width[k]*base[2];
      }
   }

}




/*** float2string *****************************************************
   Convert a float into an ascii string.
**********************************************************************/
float2string( f, str )
float f;
char *str;
{
   if (f==0.0 || fabs(f)<0.01)
      strcpy( str, "0.0" );
   else if (f>=100.0 || f<=-100.0)
      sprintf(str, "%d", (int) f );
   else
      sprintf(str, "%4.2f", f );
}
